@@ -4,17 +4,18 @@ concurrency
 intended for (potentially heavy) data processing
 """
 
-import os
-import concurrent.futures
-from typing import List, Tuple, Dict, Iterable, Callable, TypeVar, Union, Set
 
-import numpy as np
+import os
+import heapq
+import concurrent.futures
+from typing import List, Tuple, Dict, Iterable, Callable, TypeVar, Union
+
 from tqdm.std import tqdm as std_tqdm  # root for type check
 from tqdm.auto import tqdm
 from tqdm.contrib import concurrent as tqdm_concurrent
 
 from stefutil.container import group_n
-from stefutil.prettier import ca
+from stefutil.prettier import check_arg as ca
 
 
 __all__ = ['conc_map', 'batched_conc_map', 'conc_yield']
@@ -28,12 +29,12 @@ BatchedMapFn = Callable[[Tuple[List[T], int, int]], List[K]]
 
 
 def _check_conc_mode(mode: str):
-    ca.check_mismatch('Concurrency Mode', mode, ['thread', 'process'])
+    ca.assert_options('Concurrency Mode', mode, ['thread', 'process'])
 
 
 def conc_map(
-        fn: MapFn, args: Iterable[T], with_tqdm: Union[bool, Dict] = False, n_worker: int = os.cpu_count(),
-        mode: str = 'thread'
+        fn: MapFn, args: Iterable[T], with_tqdm: Union[bool, Dict] = False, n_worker: int = os.cpu_count() - 1,
+        mode: str = 'process'
 ) -> Iterable[K]:
     """
     Wrapper for `concurrent.futures.map`
@@ -97,7 +98,7 @@ class BatchedMap:
 
 def batched_conc_map(
         fn: Union[MapFn, BatchedMapFn],
-        args: Union[Iterable[T], List[T], np.array], n: int = None, n_worker: int = os.cpu_count(),
+        args: Union[Iterable[T], List[T]], n: int = None, n_worker: int = os.cpu_count(),
         batch_size: int = None,
         with_tqdm: Union[bool, dict, tqdm] = False,
         is_batched_fn: bool = False,
@@ -164,25 +165,29 @@ def batched_conc_map(
 
 
 class BatchedFn:
-    def __init__(self, fn: MapFn, pbar=None):
+    def __init__(self, fn: MapFn = None, fn_keyword: str = None, pbar=None):
         self.fn = fn
+        self.fn_keyword = fn_keyword
         self.pbar = pbar
 
-    def __call__(self, args: Iterable[T]) -> Set[K]:
+    def __call__(self, args: Iterable[T]) -> List[K]:
         """
         No order enforced
         """
-        ret = set()
+        ret = []
         for a in args:
-            ret.add(self.fn(a))
-            if self.pbar:
+            # ret.append(self.fn(a))
+            ret.append(self.fn(**{self.fn_keyword: a}) if self.fn_keyword else self.fn(a))
+            if self.pbar is not None:
                 self.pbar.update(1)
         return ret
 
 
 def conc_yield(
-        fn: MapFn, args: Iterable[T], with_tqdm: Union[bool, Dict] = False, n_worker: int = os.cpu_count(),
-        mode: str = 'thread', batch_size: Union[int, bool] = None
+        fn: MapFn, args: Iterable[T], fn_kwarg: str = None, with_tqdm: Union[bool, Dict] = False,
+        n_worker: int = os.cpu_count()-1,  # since the calling script consumes one process
+        mode: str = 'process', batch_size: Union[int, bool] = None,
+        process_chunk_size: int = None, process_chunk_multiplier: int = None, enforce_order: bool = False
 ) -> Iterable[K]:
     """
     Wrapper for `concurrent.futures`, yielding results as they become available, irrelevant of order
@@ -190,6 +195,7 @@ def conc_yield(
 
     :param fn: A function
     :param args: A list of elements as input to the function
+    :param fn_kwarg: If given, the keyword argument to pass to `fn`
     :param with_tqdm: If true, progress bar is shown
         If dict, treated as `tqdm` concurrent kwargs
             note `chunksize` is helpful
@@ -198,11 +204,18 @@ def conc_yield(
         Function has to be pickleable if 'process'
     :param batch_size: Number of elements for each sub-process worker
         Intended to lower concurrency overhead
+    :param process_chunk_size: If mode is `process`,
+        elements are chunked into groups of `process_chunk_size` for a sequence of concurrent pools
+    :param process_chunk_multiplier: If mode is `process`,
+        elements are chunked into groups of `n_worker` x `batch_size` x `process_chunk_multiplier`
+            for a sequence of concurrent pools
+    :param enforce_order: If true, results are yielded in the order of args passed
     :return: Iterator of `lst` elements mapped by `fn` with thread concurrency
     """
     _check_conc_mode(mode=mode)
 
-    cls = concurrent.futures.ThreadPoolExecutor if mode == 'thread' else concurrent.futures.ProcessPoolExecutor
+    is_thread = mode == 'thread'
+    cls = concurrent.futures.ThreadPoolExecutor if is_thread else concurrent.futures.ProcessPoolExecutor
     executor = cls(max_workers=n_worker)
 
     pbar = None
@@ -211,25 +224,97 @@ def conc_yield(
         pbar = tqdm(**tqdm_args)
 
     if batch_size:
-        batch_size = 32 if isinstance(batch_size, int) else batch_size
+        batch_size = 32 if isinstance(batch_size, bool) else batch_size
         # pbar doesn't work w/ pickle hence multiprocessing
-        fn = BatchedFn(fn=fn, pbar=pbar if mode == 'thread' else None)
-        futures = set(executor.submit(fn, args_) for args_ in group_n(args, batch_size))
-        for f in concurrent.futures.as_completed(futures):
-            if mode == 'thread':
-                yield from f.result()
-            else:
+        fn = BatchedFn(fn=fn, fn_keyword=fn_kwarg, pbar=pbar if is_thread else None)
+        if process_chunk_size is not None:
+            chunk = process_chunk_size
+        elif process_chunk_multiplier is not None:
+            chunk = n_worker * batch_size * process_chunk_multiplier
+        else:
+            chunk = None
+
+        if not is_thread and chunk:
+            it_args = group_n(args, chunk)
+            for args_ in it_args:
+                futures = {executor.submit(fn, args_): i for i, args_ in enumerate(group_n(args_, batch_size))}
+
+                pq, index = ([], 0) if enforce_order else (None, None)
+
+                for f in concurrent.futures.as_completed(futures):
+                    res = f.result()
+
+                    if enforce_order:
+                        heapq.heappush(pq, (futures[f], res))
+                        while pq and pq[0][0] == index:
+                            _, res = heapq.heappop(pq)
+                            yield from res
+                            index += 1
+
+                            if with_tqdm:
+                                pbar.update(len(res))
+
+                    else:
+                        if with_tqdm:
+                            pbar.update(len(res))
+                        yield from res
+
+                    del futures[f]
+        else:
+            futures = {executor.submit(fn, args_): i for i, args_ in enumerate(group_n(args, batch_size))}
+
+            pq, index = ([], 0) if enforce_order else (None, None)
+
+            for f in concurrent.futures.as_completed(futures):
                 res = f.result()
-                if pbar:
-                    pbar.update(len(res))
-                yield from res
+
+                if enforce_order:
+                    heapq.heappush(pq, (futures[f], res))
+                    while pq and pq[0][0] == index:
+                        _, res = heapq.heappop(pq)
+                        yield from res
+                        index += 1
+
+                        if not is_thread and with_tqdm:
+                            pbar.update(len(res))
+
+                else:
+                    if not is_thread and with_tqdm:
+                        pbar.update(len(res))
+                    yield from res
+
+                del futures[f]
+
     else:
-        futures = set(executor.submit(fn, a) for a in args)
-        for f in concurrent.futures.as_completed(futures):
+        if fn_kwarg:
+            futures = {executor.submit(fn, **{fn_kwarg: a}): i for i, a in enumerate(args)}
+        else:
+            futures = {executor.submit(fn, a): i for i, a in enumerate(args)}
+
+        if enforce_order:
+            pq = []  # priority queue to hold (index, result) pairs
+            index = 0  # the next index to yield
+        else:
+            pq, index = None, None
+
+        for f in concurrent.futures.as_completed(futures):  # TODO: process chunking
             res = f.result()
-            if with_tqdm:
-                pbar.update(1)
-            yield res
+
+            if enforce_order:
+                heapq.heappush(pq, (futures[f], res))  # prioritize by index
+                while pq and pq[0][0] == index:
+                    _, res = heapq.heappop(pq)
+                    yield res
+                    index += 1
+
+                    if with_tqdm:
+                        pbar.update(1)
+            else:
+                if with_tqdm:
+                    pbar.update(1)
+                yield res
+            
+            del futures[f]
 
 
 # DEV = True
@@ -238,30 +323,33 @@ if DEV:
     import time
     import random
 
-    from stefutil.prettier import pl, mic
+    from stefutil.prettier import style, icecream as sic, get_logger
 
+    logger = get_logger('Conc Dev')
 
-    def work(task_idx):
-        t = random.uniform(1, 4)
-        t = round(t, 3)
-        print(f'Task {pl.i(task_idx)} launched, will sleep for {pl.i(t)} s')
+    def _work(task_idx: int = None):
+        t = round(random.uniform(0.5, 3), 3)
+        print(f'Task {style(task_idx)} launched, will sleep for {style(t)} s')
         time.sleep(t)
 
-        print(f'Task {pl.i(task_idx)} is done')
+        print(f'Task {style(task_idx)} is done')
         return task_idx
 
 
+    def _dummy_fn(xx: int = None):
+        x = xx
+        t_ms = random.randint(500, 3000)
+
+        # logger.info(f'Calling dummy_fn w/ arg {s.i(x)}, will sleep for {s.i(t_ms)}ms')
+        time.sleep(t_ms / 1000)
+
+        return x, [random.random() for _ in range(int(1e6))]
+
+
 if __name__ == '__main__':
-    def work(task_idx):
-        t = random.uniform(1, 4)
-        print(f'Task {pl.i(task_idx)} launched, will sleep for {pl.i(t)} s')
-        time.sleep(t)
-
-        print(f'Task {pl.i(task_idx)} is done')
-        return task_idx
+    from stefutil.prettier import rich_progress, rich_console_log
 
     def try_concurrent_yield():
-
         # this will gather all results and return
         # with ThreadPoolExecutor() as executor:
         #     for result in executor.map(work, range(3)):
@@ -269,12 +357,21 @@ if __name__ == '__main__':
         # executor = concurrent.futures.ThreadPoolExecutor()
         executor = concurrent.futures.ProcessPoolExecutor()
         args = list(range(4))
-        futures = [executor.submit(work, a) for a in args]
+        futures = [executor.submit(_work, a) for a in args]
         for f in concurrent.futures.as_completed(futures):
             res = f.result()
-            mic(res)
+            sic(res)
     # try_concurrent_yield()
 
+    def check_conc_map():
+        n = 10
+        # for res in conc_map(fn=_work, args=range(n), with_tqdm=dict(total=n), n_worker=4, mode='process'):
+        #     sic(res)
+        from stefutil.prettier import rich_console_log
+        for res in rich_progress(conc_map(fn=_work, args=range(n), mode='process'), total=n):
+            rich_console_log(res)
+    # check_conc_map()
+
     def check_conc_yield():
         batch = False
 
@@ -291,6 +388,55 @@ if __name__ == '__main__':
 
         mode = 'process'
         # mode = 'thread'
-        for res in conc_yield(fn=work, args=it, with_tqdm=with_tqdm, n_worker=n_worker, mode=mode, batch_size=bsz):
-            mic(res)
-    check_conc_yield()
+        for res in conc_yield(fn=_work, args=it, with_tqdm=with_tqdm, n_worker=n_worker, mode=mode, batch_size=bsz):
+            sic(res)
+    # check_conc_yield()
+
+    def test_conc_mem_use():
+        from concurrent.futures import ThreadPoolExecutor, as_completed
+
+        n = 30
+        # remove_job = True
+        remove_job = False
+
+        # test_code = True
+        test_code = False
+        if test_code:
+            n_processed = 0
+            with ThreadPoolExecutor(max_workers=3) as executor:
+                futures = {executor.submit(_dummy_fn, i) for i in range(n)}
+                for f in as_completed(futures):
+                    i, _ = f.result()
+                    if remove_job:
+                        futures.remove(f)
+                    n_processed += 1
+                    logger.info(f'Process {style(i)} terminated, {style(n_processed)} / {style(n)} processed')
+        else:
+            args = dict(with_tqdm=dict(total=n), n_worker=3, mode='process', batch_size=4)
+            for i in conc_yield(fn=_dummy_fn, args=range(n), **args):
+                i, _ = i
+                logger.info(f'Process {style(i)} terminated')
+    # test_conc_mem_use()
+
+    def check_conc_process_chunk():
+        # n = 200
+        # n = 100
+        n = 50
+        # chunk_mult = 4
+        # bsz, n_worker = 4, 4
+        bsz = 4
+        # bsz = False
+        # args = dict(n_worker=4, mode='process', batch_size=bsz, process_chunk_size=chunk_mult * bsz * n_worker)
+        args = dict(n_worker=4, mode='process', batch_size=bsz, process_chunk_multiplier=None)
+
+        # for i in conc_yield(fn=_dummy_fn, fn_kwarg='xx', args=range(n), with_tqdm=dict(total=n), **args):
+        #     i, _ = i
+        #     logger.info(f'Process {s.i(i)} terminated')
+
+        lst = []
+        for i in rich_progress(conc_yield(fn=_dummy_fn, fn_kwarg='xx', args=range(n), **args, enforce_order=True), total=n):
+            i, _ = i
+            # logger.info(f'Process {s.i(i)} terminated')
+            lst.append(i)
+        rich_console_log(lst)
+    check_conc_process_chunk()
